/* 
 * update_display.c
 *
 * routines to perform the refresh of an XImage
 * they are put in a separate file due to very special ones: most of them are
 * a collection of special cases in order to get most accuracy
 *
 */

#define __UPDATE_DSP_C_

/* nonsense in linux SuperVga and ggi code */
#ifndef svgalib
#ifndef ggi

#include "xmame.h"

static void (*update_display_func)(void);
static void (*dirty_update_display_func)(void);
static char *copybuffer;

void osd_mark_dirty(int sx,int sy, int ex, int ey, int ui)
{
	int line;
	int real_width;
	char *spt,*ept;
	
	if (use_layer)
	{
	   layer_mark_rectangle_dirty_norotate(Machine->dirtylayer,sx,ex,sy,ey);
	   return;
	}
	
	if (use_dirty)
	{
	   if (sx < visual.min_x) sx=visual.min_x;
	   if (sy < visual.min_y) sy=visual.min_y;
	   if (ex > visual.max_x) ex=visual.max_x;
	   if (ey > visual.max_y) ey=visual.max_y;
	   /* calculate pointers */
	   spt = (char *)bitmap->line[sy] + sx;
	   ept = (char *)bitmap->line[sy] + ex+1;
	   /* make first visible line line 0 */
	   sy -= visual.min_y;
	   ey -= visual.min_y;
	   /* make first visible column column 0 */
	   sx -= visual.min_x;
	   /* calculate real bitmap width */
	   real_width = ((bitmap->width + 7) & ~7) + 16;
	   
	   for (line=sy; line<=ey; line++, spt+=real_width, ept+=real_width)
	   {
	       /* if start=0 this line hasn't been marked before, don't do 
	          this check for x even is x=0 the line can still be marked 
	          before (BugFix). */
	       if ( !dirty_lines[line].start )
	       {
	          dirty_lines[line].start = spt;
	          dirty_lines[line].x     = sx;
	          dirty_lines[line].end   = ept;
	       }
	       else
	       { 
	          if ( spt < dirty_lines[line].start) dirty_lines[line].start = spt;
	          if ( ept > dirty_lines[line].end  ) dirty_lines[line].end   = ept;
	          if ( sx  < dirty_lines[line].x    ) dirty_lines[line].x     = sx;
	       }
	   }
	}
}

static void merge_dirty(void)
{
   struct dirty_line *temp;
   int i;
   
   /* merge this frame's dirty lines, with the last one's (old),
      store the result in the old one's since these aren't needed
      anymore */
      
   for (i=0; i<visual_height; i++)
      /* no need to do anything if current line not dirty */
      if (dirty_lines[i].start)
      {
         /* both current and old line are dirty */
         if(old_dirty_lines[i].start)
         {
            if (dirty_lines[i].start < old_dirty_lines[i].start)
            {
               old_dirty_lines[i].start = dirty_lines[i].start;
               old_dirty_lines[i].x     = dirty_lines[i].x;
            }
            if (dirty_lines[i].end   > old_dirty_lines[i].end)
               old_dirty_lines[i].end   = dirty_lines[i].end;
         }
         else
            old_dirty_lines[i]=dirty_lines[i];
      }
   
   /* now swap dirty_buffers */
   temp            = old_dirty_lines;
   old_dirty_lines = dirty_lines;
   dirty_lines     = temp;
}

extern short layer_dirty_shift;

/* copy the dirty info in the layer system to our own dirty_lines */
void copy_dirty_layer_to_dirty(void)
{
   int i,j;
   int block, start, end = 0;
   char *spt = (char *)bitmap->line[visual.min_y];
   /* calculate real bitmap width */
   int real_width = ((bitmap->width + 7) & ~7) + 16;
   
   for (i=0; i<visual_height; i++, spt+=real_width)
   {
      block = ((i + visual.min_y) >> 3) << layer_dirty_shift;
      for (j=0, start=-1; j<bitmap->width; j+=8, block++)
      {
         if (Machine->dirtylayer->dirty[block])
         {
           if (start == -1)
           {
              start = j;
              end   = j;
           }
           else
              end   = j;
         }
      }
      
      if (start != -1)
      {
         end += 8;
         if (start < visual.min_x) start = visual.min_x;
         if (end   > visual.max_x) end   = visual.max_x;
         dirty_lines[i].start = spt + start;
         dirty_lines[i].end   = spt + end;
         dirty_lines[i].x     = start - visual.min_x;;
      }
   }
   
   layer_mark_all_dirty(Machine->dirtylayer, MARK_ALL_CLEAN);
}

/* used when expose events received */
void osd_refresh_screen(void)
{

#ifdef USE_MITSHM
    if (mit_shm_avail)
    {
	XShmPutImage (display, window, gc, image, 0, 0, 0, 0, image->width, image->height, FALSE);
    } 
    else /* no shm: just use XPutImage */
#endif
        XPutImage (display, window, gc, image, 0, 0, 0, 0, image->width, image->height);

   /* some games "flickers" with XFlush, so command line option is provided */
   if(use_xsync) XSync (display,False); /* be sure to get request processed */
   else XFlush (display); 		/* flush buffer to server */
}

void osd_refresh_dirty_screen(void) {
    int line,lineto;
#ifdef USE_MITSHM
    if (mit_shm_avail) {
	/* just parse every dirty lines and send it to X-Server */
	for (line=0; line<visual_height; line++) {
	    if (!dirty_lines[line].start) continue;
	    /* go till next non-dirty line */
	    for(lineto=line; dirty_lines[lineto].start; lineto++);
	    XShmPutImage (display, window, gc, image, 
			0, line*heightscale, 0, line*heightscale , 
			image->width, (lineto-line)*heightscale, FALSE);
	    /* now mark parsed lines as clean */
	    memset((void *)&dirty_lines[line], 0, (lineto-line)*sizeof(struct dirty_line) );
	    line=lineto;
	}
    } else { /* no shm: just use XPutImage */
#endif
	for (line=0; line<visual_height; line++) {
	    if (!dirty_lines[line].start) continue; 
	    /* go till next non-dirty line */
	    for(lineto=line; dirty_lines[lineto].start; lineto++);
            XPutImage (display, window, gc, image, 
			0, line*heightscale, 0, line*heightscale, 
			image->width, (lineto-line)*heightscale);
	    /* now mark parsed lines as clean */
	   memset((void *)&dirty_lines[line], 0, (lineto-line)*sizeof(struct dirty_line) );
	   line=lineto;
	}
#ifdef USE_MITSHM
    }
#endif

   /* some games "flickers" with XFlush, so command line option is provided */
   if(use_xsync) XSync (display,False); /* be sure to get request processed */
   else XFlush (display); 		/* flush buffer to server */
}

/* invoked by main tree code to update bitmap into screen */
void sysdep_update_display(void)
{
    int old_use_dirty=use_dirty;
    int old_use_layer=use_layer;
    
    /* if F12 pressed, make a snapshot of the screen */
    if (osd_key_pressed(OSD_KEY_F12) ) osd_snapshot();
    
    if (use_dirty && !use_layer) merge_dirty();
    if (use_dirty &&  use_layer) copy_dirty_layer_to_dirty();
    
    if (palette_dirty)
    {
       /* trick make this a dirty update and mark everything dirty, to make
          sure the entire screen is updated since the palette has changed */
       use_dirty=TRUE;
       use_layer=FALSE;
       osd_mark_dirty(0, 0, bitmap->width - 1, bitmap->height - 1, 0);
       /* update copybuffer if we're normally not using dirty code */
       if (old_use_dirty==0) 
       {
          int line;
          char *dst_pt = copybuffer;
          for (line=visual.min_y; line<visual.max_y; line++, dst_pt+=visual_width)
              memcpy( (void*)dst_pt, (void *)(bitmap->line[line]+visual.min_x), visual_width ); 
       }
    }
    
    /* call propper update routine */
    if (use_dirty)
       (*dirty_update_display_func)();
     else
       (*update_display_func)();
       
    use_dirty=old_use_dirty;
    use_layer=old_use_layer;
    palette_dirty=0;
}

void do_update_pseudo_dirty_scale1(void) {
	char *src_pt, *dest_pt;
	int line;
	for (line=0; line<visual_height; line++)
	{
	   dest_pt = scaled_buffer_ptr + line*visual_width + dirty_lines[line].x;
	   for (src_pt=dirty_lines[line].start; src_pt<dirty_lines[line].end; src_pt++, dest_pt++)
	      *dest_pt = (char)xpixel[0xff & *src_pt];
	}   
	osd_refresh_dirty_screen();
}

void do_update_pseudo_scale1(void) {
   long *end, *spt;
   long *dpt   =(long *)copybuffer;
   char *dpypt = scaled_buffer_ptr;
   int line, update=0;
   for (line=visual.min_y; line<=visual.max_y; line++)
   {
      spt=(long *)( (char *)bitmap->line[line] + visual.min_x);
      end=(long *)( (char *)bitmap->line[line] + visual.min_x + visual_width );
      for (;spt<end;spt++,dpt++,dpypt+=sizeof(long))
      {
         if (*spt!=*dpt) {
            *dpypt    =xpixel[0xff & *((char *)spt  )];
            *(dpypt+1)=xpixel[0xff & *((char *)spt+1)];
            *(dpypt+2)=xpixel[0xff & *((char *)spt+2)];
            *(dpypt+3)=xpixel[0xff & *((char *)spt+3)];
#ifdef linux_alpha  /* 64bit machines, */
            *(dpypt+4)=xpixel[0xff & *((char *)spt+4)];
            *(dpypt+5)=xpixel[0xff & *((char *)spt+5)];
            *(dpypt+6)=xpixel[0xff & *((char *)spt+6)];
            *(dpypt+7)=xpixel[0xff & *((char *)spt+7)];
#endif
            *dpt=*spt;
            update++;
         }
      }
   }
   if (update) osd_refresh_screen();
}

void do_update_pseudo_generic(void) {
   int i, line;
   int x,y,px,py;
   long color;
   char *dpypt;
   long *spt, *end;
   long *dpt=(long *)copybuffer;
   for (line=visual.min_y, y=0; line<=visual.max_y; line++, y++)
   {
      spt=(long *)( (char *)bitmap->line[line] + visual.min_x);
      end=(long *)( (char *)bitmap->line[line] + visual.min_x + visual_width );
      for (x=0;spt<end;spt++,dpt++,x+=sizeof(long)) {
         if (*spt!=*dpt) {
            for (i=0;i<sizeof(long);i++)
            {
               if ( *((char *)spt+i) == *((char *)dpt+i) ) continue;
               /* now perform a byte-by-byte search in current long pt */
               color=xpixel[0xff & *((char *)spt+i)];
               dpypt=scaled_buffer_ptr + heightscale*y*image->bytes_per_line + (x+i)*widthscale;
               for(py=0;py<heightscale;py++) 
                  for(px=0;px<widthscale;px++) 
                     *(dpypt+py*image->bytes_per_line+px)=color;
            } /* for i */
            *dpt=*spt;
         } /* if long is not equal */
      } /* for buffer */
   } /* for line */
   osd_refresh_screen();
}

void do_update_pseudo_dirty_generic(void) {
    char *pt;
    char *dpt,*auxpt;
    int line;
    int i;
    for (line=0; line<visual_height; line++) {
       if( ! dirty_lines[line].start ) continue;
       dpt = auxpt = scaled_buffer_ptr + heightscale*line*image->bytes_per_line + widthscale*dirty_lines[line].x;
       for(pt=dirty_lines[line].start;pt<dirty_lines[line].end;pt++,dpt+=widthscale)
	   for(i=0;i<widthscale;i++) *(dpt+i)=(char)xpixel[0xff & *pt];
       /* and now perform a memcpy to next lines */
       for (i=1;i<heightscale;i++)
           memcpy(auxpt+i*image->bytes_per_line,auxpt,widthscale*(dirty_lines[line].end-dirty_lines[line].start));
    }
    osd_refresh_dirty_screen();
}

void do_update_pseudo_dirty_scale2(void) {
    char *pt;
    char *dpt,*auxpt;
    int line;
    for (line=0; line<visual_height; line++) {
       if( ! dirty_lines[line].start ) continue;
       dpt = auxpt =
	    scaled_buffer_ptr + 2*line*image->bytes_per_line + 2*dirty_lines[line].x;
       for(pt=dirty_lines[line].start;pt<dirty_lines[line].end;pt++,dpt+=2) {
	   *dpt    =(char)xpixel[0xff & *pt];
	   *(dpt+1)=(char)xpixel[0xff & *pt];
       }
       /* and now perform a memcpy to next line */
       memcpy(auxpt+image->bytes_per_line,auxpt,2*(dirty_lines[line].end-dirty_lines[line].start));
    }
    osd_refresh_dirty_screen();
}

void do_update_pseudo_scale2(void) {
   int i,line;
   int x,y;
   long color;
   char *dpypt;
   long *spt, *end;
   long *dpt=(long *)copybuffer;
   for (line=visual.min_y, y=0; line<=visual.max_y; line++, y++)
   {
      spt=(long *)( (char *)bitmap->line[line] + visual.min_x);
      end=(long *)( (char *)bitmap->line[line] + visual.min_x + visual_width );
      for (x=0;spt<end;spt++,dpt++,x+=sizeof(long)) {
         if (*spt!=*dpt) {
            for (i=0;i<sizeof(long);i++) {
               if ( *((char *)spt+i) == *((char *)dpt+i) ) continue;
               /* now perform a byte-by-byte search in current long pt */
               color=xpixel[0xff & *((char *)spt+i)];
               dpypt=scaled_buffer_ptr + ((y*image->bytes_per_line + x + i) << 1);
               /* and draw pixel in image */
               *dpypt=color;*(dpypt+1)=color; 
               dpypt+=image->bytes_per_line;
               *dpypt=color; *(dpypt+1)=color;
            } /* for i */
            *dpt=*spt; /* mark long as updated */
         } /* if long is not equal */
      } /* for buffer */
   } /* for line */
   osd_refresh_screen();
}

void do_update_pseudo_dirty_scale3(void)
{
    char *pt;
    char *dpt,*auxpt;
    int line;
    for (line=0; line<visual_height; line++) {
       if( ! dirty_lines[line].start ) continue;
       dpt = auxpt =
	    scaled_buffer_ptr + 3*line*image->bytes_per_line + 3*dirty_lines[line].x;
       for(pt=dirty_lines[line].start;pt<dirty_lines[line].end;pt++,dpt+=3) {
	   *dpt    =(char)xpixel[0xff & *pt];
	   *(dpt+1)=(char)xpixel[0xff & *pt];
	   *(dpt+2)=(char)xpixel[0xff & *pt];
       }
       /* and now perform a memcpy to next line */
       memcpy(auxpt+image->bytes_per_line,auxpt,3*(dirty_lines[line].end-dirty_lines[line].start));
       memcpy(auxpt+2*image->bytes_per_line,auxpt,3*(dirty_lines[line].end-dirty_lines[line].start));
    }
    osd_refresh_dirty_screen();
}

void do_update_pseudo_scale3(void)
{
   int i,line;
   int x,y;
   long color;
   char *dpypt;
   long *spt, *end;
   long *dpt=(long *)copybuffer;
   for (line=visual.min_y, y=0; line<=visual.max_y; line++, y++)
   {
      spt=(long *)( (char *)bitmap->line[line] + visual.min_x);
      end=(long *)( (char *)bitmap->line[line] + visual.min_x + visual_width );
      for (x=0;spt<end;spt++,dpt++,x+=sizeof(long)) {
         if (*spt!=*dpt) {
            for (i=0;i<sizeof(long);i++) {
               if ( *((char *)spt+i) == *((char *)dpt+i) ) continue;
               /* now perform a byte-by-byte search in current long pt */
               color=xpixel[0xff & *((char *)spt+i)];
               dpypt=scaled_buffer_ptr + 3*(y*image->bytes_per_line + x + i);
               /* and draw pixel in image */
               *dpypt=color;*(dpypt+1)=color; *(dpypt+2)=color;
               dpypt+=image->bytes_per_line;
               *dpypt=color;*(dpypt+1)=color; *(dpypt+2)=color;
               dpypt+=image->bytes_per_line;
               *dpypt=color;*(dpypt+1)=color; *(dpypt+2)=color;
            } /* for i */
            *dpt=*spt; /* mark long as updated */
         } /* if long is not equal */
      } /* for buffer */
   } /* for line */
   osd_refresh_screen();
}

void do_update_truec_dirty_generic(void)
{
	char *pt;
	int x,y;
	int i,j;
	int line;
	long color;
	for (y=0, line=0; line<visual_height; line++, y+=heightscale) {
	   if (!dirty_lines[line].start) continue;
	   x = dirty_lines[line].x * widthscale;
	   for (pt=dirty_lines[line].start; pt<dirty_lines[line].end; pt++, x+=widthscale)
	   {
	      color=xpixel[0xff & *pt];
	      for(j=0;j<heightscale;j++) 
	         for(i=0;i<widthscale;i++) XPutPixel(image,x+i,y+j,color);
	   }
	}
	osd_refresh_dirty_screen();
}

void do_update_truec_generic(void)
{
   int i, line;
   int tmp_x,x,y,px,py;
   long color;
   long *spt, *end;
   long *dpt=(long *)copybuffer;
   for (line=visual.min_y, y=0; line<=visual.max_y; line++, y+=heightscale)
   {
      spt=(long *)( (char *)bitmap->line[line] + visual.min_x);
      end=(long *)( (char *)bitmap->line[line] + visual.min_x + visual_width );
      for (x=0;spt<end;spt++,dpt++,x+=sizeof(long)*widthscale) {
         if (*spt!=*dpt) {
            for (tmp_x=x,i=0;i<sizeof(long);i++,tmp_x+=widthscale)
            {
               if ( *((char *)spt+i) == *((char *)dpt+i) ) continue;
               /* now perform a byte-by-byte search in current long pt */
               color=xpixel[0xff & *((char *)spt+i)];
               for(py=0;py<heightscale;py++) 
                  for(px=0;px<widthscale;px++)
                     XPutPixel(image,tmp_x+px,y+py,color);
            } /* for i */
            *dpt=*spt;
         } /* if long is not equal */
      } /* for buffer */
   } /* for line */
   osd_refresh_screen();
}

void do_update_truec_dirty_scale1(void)
{
	char *pt;
	int line,x;

	for (line=0; line<visual_height; line++) {
	   if (!dirty_lines[line].start) continue;
	   x = dirty_lines[line].x;
	   for (pt=dirty_lines[line].start; pt<dirty_lines[line].end; pt++,x++)
		XPutPixel(image,x,line,xpixel[0xff & *pt]);
	}
	osd_refresh_dirty_screen();
}

void do_update_truec_scale1(void)
{
   int x,y;
   int line, update=0;
   long *end, *spt;
   long *dpt   =(long *)copybuffer;
   for (line=visual.min_y, y=0; line<=visual.max_y; line++, y++)
   {
      spt=(long *)( (char *)bitmap->line[line] + visual.min_x);
      end=(long *)( (char *)bitmap->line[line] + visual.min_x + visual_width );
      for (x=0;spt<end;spt++,dpt++,x+=sizeof(long))
      {
         if (*spt!=*dpt) {
            XPutPixel(image,x  ,y,xpixel[0xff & *((char *)spt  )]);
            XPutPixel(image,x+1,y,xpixel[0xff & *((char *)spt+1)]);
            XPutPixel(image,x+2,y,xpixel[0xff & *((char *)spt+2)]);
            XPutPixel(image,x+3,y,xpixel[0xff & *((char *)spt+3)]);
#ifdef linux_alpha  /* 64bit machines, */
            XPutPixel(image,x+4,y,xpixel[0xff & *((char *)spt+4)]);
            XPutPixel(image,x+5,y,xpixel[0xff & *((char *)spt+5)]);
            XPutPixel(image,x+6,y,xpixel[0xff & *((char *)spt+6)]);
            XPutPixel(image,x+7,y,xpixel[0xff & *((char *)spt+7)]);
#endif
            *dpt=*spt;
            update++;
         }
      }
   }
   if (update) osd_refresh_screen();
}

void do_update_truec_dirty_scale2(void)
{
	char *pt;
	int x,y;
	int line;
	long color;
	for (line=0, y=0; line<visual_height; line++, y+=2) {
	   if (!dirty_lines[line].start) continue;
	   x = dirty_lines[line].x << 1;
	   for (pt=dirty_lines[line].start; pt<dirty_lines[line].end; pt++,x+=2)
	   {
	      color=xpixel[0xff & *pt];
	      XPutPixel(image,x,y,color); XPutPixel(image,x+1,y,color);
	      XPutPixel(image,x,y+1,color); XPutPixel(image,x+1,y+1,color);
	   }
	}
	osd_refresh_dirty_screen();
}

void do_update_truec_scale2(void)
{
   int i,line;
   int x,y;
   long color;
   long *spt, *end;
   long *dpt=(long *)copybuffer;
   for (line=visual.min_y, y=0; line<=visual.max_y; line++, y+=2)
   {
      spt=(long *)( (char *)bitmap->line[line] + visual.min_x);
      end=(long *)( (char *)bitmap->line[line] + visual.min_x + visual_width );
      for (x=0;spt<end;spt++,dpt++,x+=sizeof(long)) {
         if (*spt!=*dpt) {
            for (i=0;i<sizeof(long);i++) {
               if ( *((char *)spt+i) == *((char *)dpt+i) ) continue;
               /* now perform a byte-by-byte search in current long pt */
               color=xpixel[0xff & *((char *)spt+i)];
               XPutPixel(image, ((x+i)<<1),   y  , color);
               XPutPixel(image, ((x+i)<<1)+1, y  , color);
               XPutPixel(image, ((x+i)<<1),   y+1, color);
               XPutPixel(image, ((x+i)<<1)+1, y+1, color);
            } /* for i */
            *dpt=*spt; /* mark long as updated */
         } /* if long is not equal */
      } /* for buffer */
   } /* for line */
   osd_refresh_screen();
}

void do_update_truec_dirty_scale3(void)
{
	char *pt;
	int x,y;
	int line;
	long color;
	for (line=0, y=0; line<visual_height; line++, y+=3) {
	   if (!dirty_lines[line].start) continue;
	   x = dirty_lines[line].x * 3;
	   for (pt=dirty_lines[line].start; pt<dirty_lines[line].end; pt++,x+=3)
	   {
	      color = xpixel[0xff & *pt];
	      XPutPixel(image,x,y  ,color); XPutPixel(image,x+1,y  ,color); XPutPixel(image,x+2,y  ,color);
	      XPutPixel(image,x,y+1,color); XPutPixel(image,x+1,y+1,color); XPutPixel(image,x+2,y+1,color);
	      XPutPixel(image,x,y+2,color); XPutPixel(image,x+1,y+2,color); XPutPixel(image,x+2,y+2,color);
	   }
	}
	osd_refresh_dirty_screen();
}

void do_update_truec_scale3(void)
{
   int i,line;
   int tmp_x,x,y;
   long color;
   long *spt, *end;
   long *dpt=(long *)copybuffer;
   for (line=visual.min_y, y=0; line<=visual.max_y; line++, y+=3)
   {
      spt=(long *)( (char *)bitmap->line[line] + visual.min_x);
      end=(long *)( (char *)bitmap->line[line] + visual.min_x + visual_width );
      for (x=0;spt<end;spt++,dpt++,x+=sizeof(long)*3) {
         if (*spt!=*dpt) {
            for (tmp_x=x,i=0; i<sizeof(long); i++,tmp_x+=3) {
               if ( *((char *)spt+i) == *((char *)dpt+i) ) continue;
               /* now perform a byte-by-byte search in current long pt */
               color=xpixel[0xff & *((char *)spt+i)];
               XPutPixel(image,tmp_x,y,  color); XPutPixel(image,tmp_x+1,y  ,color); XPutPixel(image,tmp_x+2,y  ,color);
               XPutPixel(image,tmp_x,y+1,color); XPutPixel(image,tmp_x+1,y+1,color); XPutPixel(image,tmp_x+2,y+1,color);
               XPutPixel(image,tmp_x,y+2,color); XPutPixel(image,tmp_x+1,y+2,color); XPutPixel(image,tmp_x+2,y+2,color);
            } /* for i */
            *dpt=*spt; /* mark long as updated */
         } /* if long is not equal */
      } /* for buffer */
   } /* for line */
   osd_refresh_screen();
}

#define DSP_SCALE_GENERIC 	0
#define DSP_SCALE_1	      	1
#define DSP_SCALE_2		2
#define DSP_SCALE_3		3

#define DSP_USE_TRUECOLOR	4
#define DSP_USE_DIRTY    	8

void (*update_funcs[])(void) = {
/* normal ( parsing image bitmap strategy */
	 do_update_pseudo_generic,
	 do_update_pseudo_scale1,
	 do_update_pseudo_scale2,
	 do_update_pseudo_scale3,
	 do_update_truec_generic,
	 do_update_truec_scale1,
	 do_update_truec_scale2,
	 do_update_truec_scale3,
/* dirty_lines strategy */
	 do_update_pseudo_dirty_generic,
	 do_update_pseudo_dirty_scale1,
	 do_update_pseudo_dirty_scale2,
	 do_update_pseudo_dirty_scale3,
	 do_update_truec_dirty_generic,
	 do_update_truec_dirty_scale1,
	 do_update_truec_dirty_scale2,
	 do_update_truec_dirty_scale3,
};

int eval_update_display_func(void)
{
	int index=0;
	
	if (force_truecolor)	     index |= DSP_USE_TRUECOLOR;
	if (heightscale!=widthscale) index |= DSP_SCALE_GENERIC;
	else {
	    if (heightscale > 3)     index |= DSP_SCALE_GENERIC;
	    else                     index |= heightscale;
	}
	
	/* now set update_display pointers */
	update_display_func       = update_funcs[index];
	dirty_update_display_func = update_funcs[index+DSP_USE_DIRTY];
	
	if (!use_dirty)
	{
	   copybuffer=(char *)malloc( visual_width * visual_height * sizeof(char) );
	   if (!copybuffer)
	   {
	      fprintf(stderr,"Cannot allocate space for copybuffer\n");
	      return OSD_NOT_OK;
	   }
	}
	
	return OSD_OK;
}

/* if not defined svgalib */
#endif
#endif /* not defined ggi */
