/*
 * X-Mame generic video code
 *
 */
#define __VIDEO_C_

#include "xmame.h"

/* Create a bitmap. Also calls osd_clearbitmap() to appropriately initialize */
/* it to the background color. */
/* VERY IMPORTANT: the function must allocate also a "safety area" 8 pixels wide all */
/* around the bitmap. This is required because, for performance reasons, some graphic */
/* routines don't clip at boundaries of the bitmap. */
struct osd_bitmap *osd_new_bitmap(int width,int height,int depth)       /* ASG 980209 */
{
	struct osd_bitmap *bitmap;

	if (!width) width++;
	if (!height) height++;

	if (Machine->orientation & ORIENTATION_SWAP_XY)
	{
		int temp;

		temp = width;
		width = height;
		height = temp;
	}

	if ((bitmap = malloc(sizeof(struct osd_bitmap))) != 0)
	{
		int i,rowlen,rdwidth;
		unsigned char *bm;
		int safety;


		if (width > 32) safety = 8;
		else safety = 0;        /* don't create the safety area for GfxElement bitmaps */

		if (depth != 8 && depth != 16) depth = 8;

		bitmap->depth = depth;
		bitmap->width = width;
		bitmap->height = height;

		rdwidth = (width + 7) & ~7;     /* round width to a quadword */
		if (depth == 16)
			rowlen = 2 * (rdwidth + 2 * safety) * sizeof(unsigned char);
		else
			rowlen =     (rdwidth + 2 * safety) * sizeof(unsigned char);

		if ((bm = malloc((height + 2 * safety) * rowlen)) == 0)
		{
			free(bitmap);
			return 0;
		}

		if ((bitmap->line = malloc(height * sizeof(unsigned char *))) == 0)
		{
			free(bm);
			free(bitmap);
			return 0;
		}

		for (i = 0;i < height;i++)
			bitmap->line[i] = &bm[(i + safety) * rowlen + safety];

		bitmap->_private = bm;

		osd_clearbitmap(bitmap);
	}

	return bitmap;
}

void osd_free_bitmap(struct osd_bitmap *bitmap)
{
	if (bitmap)
	{
		free(bitmap->line);
		free(bitmap->_private);
		free(bitmap);
		bitmap = NULL;
	}
}

/* set the bitmap to black */
void osd_clearbitmap(struct osd_bitmap *bitmap)
{
	int i;
#ifdef MESS
	extern int scrbitmap_dirty;
	scrbitmap_dirty = 1;
#endif

	for (i = 0;i < bitmap->height;i++)
	{
		if (bitmap->depth == 16)
			memset(bitmap->line[i],0,2*bitmap->width);
		else
			memset(bitmap->line[i],black_pen,bitmap->width);
	}

	if (bitmap == Machine->scrbitmap)
	{
		osd_mark_dirty (0,0,bitmap->width-1,bitmap->height-1,1);

		/* signal the layer system that the screenneeds a complete refresh */
		layer_mark_full_screen_dirty();
	}
}

struct osd_bitmap *osd_create_display(int width, int height, int attributes)
{
   int a;

   bitmap = osd_create_bitmap(width,height);
   if (bitmap==NULL) return NULL;
   
#ifndef MESS
   if (Machine->drv->video_attributes & VIDEO_TYPE_VECTOR)
   {
      visual.min_x = 0;
      visual.min_y = 0;
      visual.max_x = width - 1;
      visual.max_y = height - 1;
   }
   else
#endif
     visual = Machine->drv->visible_area;
   
#ifndef MESS
   if( Machine->orientation & ORIENTATION_SWAP_XY) {
      a=visual.max_x; visual.max_x=visual.max_y; visual.max_y=a;
      a=visual.min_x; visual.min_x=visual.min_y; visual.min_y=a;
      a=width; width=height; height=a;
   } 
#endif

   if (Machine->orientation & ORIENTATION_FLIP_X)
   {
      a            = width - visual.min_x - 1;
      visual.min_x = width - visual.max_x - 1;
      visual.max_x = a;
   }
   
   if (Machine->orientation & ORIENTATION_FLIP_Y)
   {
      a            = height - visual.min_y - 1;
      visual.min_y = height - visual.max_y - 1;
      visual.max_y = a;
   }
   
   visual_width = visual.max_x - visual.min_x + 1;
   visual_height = visual.max_y - visual.min_y + 1;

   /* Can we do dirty? First check if the driver uses the GfxLayer system */
   if (Machine->drv->layer)
   {
      use_layer = TRUE;
   }
   else if ( (Machine->drv->video_attributes & VIDEO_SUPPORTS_DIRTY) == 0 )
   {
      use_dirty = FALSE;
   }
   
   if (sysdep_create_display()==OSD_NOT_OK)
   {
      osd_free_bitmap(bitmap);
      return NULL;
   }
   
   set_ui_visarea (visual.min_x, visual.min_y, visual.max_x, visual.max_y);
   
   /* Clear the current palette */
   memset((void *)current_palette, 0 , 3*256*sizeof(unsigned char));
   
   /* for debugging only */
   printf ("height = %d, widt= %d, visheight = %d, viswidth = %d,"
           "visstarty= %d, visstartx= %d\n",
            height, width, visual_height, visual_width, visual.min_y,
            visual.min_x);
   
   return bitmap;
}   

void osd_allocate_colors(unsigned int totalcolors,const unsigned char *palette,unsigned short *pens)
{
   int a;

   if (totalcolors>256)
   {
      fprintf(stderr, "Warning: More than 256 colors (%d) are needed for this emulation\n", totalcolors);
      /* fill the remainder of the pens array with 0's to make sure */
      /* nothing strange happens                                    */
      for (a=256;a<totalcolors;a++) pens[a]=0;
      totalcolors=256;
   }
   
   /* init the palette */
   sysdep_init_palette(totalcolors, palette, pens);
   
   /* find black_pen */
   for(a=0;a<256;a++)
      if (current_palette[3*a   ] == 0 &&
          current_palette[3*a +1] == 0 &&
          current_palette[3*a +2] == 0 )
      {
         black_pen = a;
         break;
      }
       
   if (black_pen != 0)
   {
      osd_clearbitmap(bitmap);
      sysdep_clear_screen();
   }
   
   palette_dirty = TRUE;
}

void osd_get_pen(int pen,unsigned char *red, unsigned char *green, unsigned char *blue)
{
    *red   = current_palette[pen*3   ];
    *green = current_palette[pen*3 +1];
    *blue  = current_palette[pen*3 +2];
}

#ifdef HAVE_GETTIMEOFDAY
/* Standard UNIX clock() is based on CPU time, not real time.
   Here is a real-time drop in replacement for UNIX systems that have the
   gettimeofday() routine.  This results in much more accurate timing for
   throttled emulation.
*/
clock_t clock()
{
  static long init_sec = 0;

  struct timeval tv;
  gettimeofday(&tv, 0);
  if (init_sec == 0) init_sec = tv.tv_sec;
  return (tv.tv_sec - init_sec) * 1000000 + tv.tv_usec;
}
#define UCLOCKS_PER_SEC 1000000
#else
#define UCLOCKS_PER_SEC CLOCKS_PER_SEC
#endif

void osd_update_display(void)
{
	static int showfps,showfpstemp,f8pressed,f10pressed,f11pressed;
	clock_t curr;
	#define MEMORY 10
	static clock_t prev[MEMORY];
	static int memory,speed;
	extern int frameskip;
	int i, need_to_clear_bitmap = 0;

	if (osd_key_pressed(OSD_KEY_F8))
	{
		if (f8pressed == 0)
		{
			frameskip = (frameskip + 1) % 4;
			showfpstemp = 50;
		}
		f8pressed = 1;
	}
	else f8pressed = 0;

	if (osd_key_pressed(OSD_KEY_F10))
	{
		if (f10pressed == 0) throttle ^= 1;
		f10pressed = 1;
	}
	else f10pressed = 0;

	if (osd_key_pressed(OSD_KEY_F11))
	{
		if (f11pressed == 0)
		{
			showfps ^= 1;
			if (showfps == 0)
			{
				need_to_clear_bitmap = 1;
			}
		}
		f11pressed = 1;
	}
	else f11pressed = 0;

	if (showfpstemp)         /* MAURY_BEGIN: nuove opzioni */
	{
		showfpstemp--;
		if ((showfps == 0) && (showfpstemp == 0))
		{
			need_to_clear_bitmap = 1;
		}
	}

	/* now wait until it's time to update the screen */
	if (throttle)
	{
		do
		{
			curr = clock();
		} while ((curr - prev[memory]) < (frameskip+1) * UCLOCKS_PER_SEC/Machine->drv->frames_per_second);
	}
	else curr = clock();

	memory = (memory+1) % MEMORY;

	if (curr - prev[memory])
	{
		int divdr;


		divdr = Machine->drv->frames_per_second * (curr - prev[memory]) / (100 * MEMORY);
		speed = (UCLOCKS_PER_SEC * (frameskip+1) + divdr/2) / divdr;
	}

	prev[memory] = curr;

	if (showfps || showfpstemp) /* MAURY: nuove opzioni */
	{
		int trueorientation;
		int fps,i,l;
		char buf[30];
		
		/* hack: force the display into standard orientation to avoid */
		/* rotating the text */
		trueorientation = Machine->orientation;
		Machine->orientation = ORIENTATION_DEFAULT;

		fps = (Machine->drv->frames_per_second / (frameskip+1) * speed + 50) / 100;
		sprintf(buf," %3d%%(%3d/%d fps)",speed,fps,Machine->drv->frames_per_second);
		l = strlen(buf);
		for (i = 0;i < l;i++)
			drawgfx(Machine->scrbitmap,Machine->uifont,buf[i],DT_COLOR_WHITE,0,0,visual.max_x-(l-i)*Machine->uifont->width,visual.min_y,0,TRANSPARENCY_NONE,0);
		Machine->orientation = trueorientation;
	}
	
	if (!black_pen_valid)
	{
		for(i=0;i<256;i++)
			if (current_palette[3*i   ] == 0 &&
			    current_palette[3*i +1] == 0 &&
			    current_palette[3*i +2] == 0 )
			{
				black_pen = i;
				sysdep_clear_screen();
				black_pen_valid = TRUE;
				break;
			}
	}
	
	sysdep_update_display();
	
	if (need_to_clear_bitmap)
		osd_clearbitmap(Machine->scrbitmap);
}
