/*
 * X-Mame video specifics code
 *
 */
#define __XDEP_C_

#ifndef svgalib
#ifndef ggi

/*
 * Include files.
 */

#include "xmame.h"
#include "sound.h" /* for rearming the timer */
#include <X11/cursorfont.h>

static Cursor   cursor;
static XVisualInfo	 myvisual;

/*
 * makes a snapshot of screen when F12 is pressed.
 *
 * store it in xpm pixmap format
 * save in users'home dir as gamename-XXX.xpm ( XXX equals to #snapshot )
 * 
 */
int osd_snapshot(void)
{
#ifdef HAS_XPM
	char name[1024];
	FILE *f;
	int res;
	XpmAttributes *xpmattr;
	do
	{
		sprintf(name,"%s/%s-snap%03d.xpm",getenv("HOME"),Machine->gamedrv->name,snapshot_no);
		/* avoid overwriting of existing files */
		if ((f = fopen(name,"rb")) != 0)
		{
			fclose(f);
			snapshot_no++;
		}
	} while (f != 0);
	/* set propper colormap */
	xpmattr=calloc(1,sizeof(XpmAttributes));
	if ( ! xpmattr ) { 
		fprintf(stderr,"calloc() Failed in osd_snapshot\n");
		return (OSD_NOT_OK);
	}
	xpmattr->valuemask	= XpmColormap;
	xpmattr->colormap	= colormap;	/* set my own colormap */
	res = XpmWriteFileFromImage(display,name,image,0,xpmattr);
	free(xpmattr);				/* cleanup task */
	if ( res != XpmSuccess ) {
		fprintf(stderr,"Failed to save snapshot %s\n",name);
		return (OSD_NOT_OK);
	} else {
		fprintf(stderr,"Saved snapshot as %s\n",name);
		/* wait until key released */
		while( osd_key_pressed(OSD_KEY_F12) );
		return (OSD_OK);
	}
#else
	fprintf(stderr,"Sorry: XPM library not included. Cannot make snapshot\n");
	return (OSD_NOT_OK);
#endif
}

/*
 * Create a display screen, or window, large enough to accomodate a bitmap
 * of the given dimensions. 
 *
 * Added: let osd_create_display allocate the actual display buffer. This
 * seems a bit dirty, but is more or less essential for X implementations
 * to avoid a lengthy memcpy().
 */

/* following routine traps missused MIT-SHM if not available */
int test_mit_shm(Display *display, XErrorEvent *error) {
	char msg[256];
	unsigned char ret=error->error_code;
	XGetErrorText(display,ret,msg,256);
	/* if MIT-SHM request failed, note and continue */
	if (error->error_code == BadAccess ) { mit_shm_avail=0; return 0; }
	/* else unspected error code: notify and exit */
	fprintf(stderr,"Unspected X Error %d: %s\n",ret,msg );
	exit(1);
	return 0; /* DCC blames... */
}

/* Create display */
int sysdep_create_display(void)
{
	XSetWindowAttributes winattr;
	XGCValues	 xgcv;
	int 		 myscreen;
	XEvent		 event;
  	XSizeHints 	 hints;
  	XWMHints 	 wm_hints;
	Status		 result;
	int		 xsize,ysize;
	int              i,j,k;
	extern char 	 *title;
	
	if (!use_dirty)
	{
	   /* round to long, otherwise non-aligned access may happen,
	   since nodirty uses long compare. */
	   visual.min_x &= ~(sizeof(long)-1);
	   visual.max_x  = ( (visual.max_x + 1) & ~(sizeof(long)-1) ) - 1;
	   /* now we need to recalculate the width */
	   visual_width  = visual.max_x - visual.min_x + 1;
	}

	/* init dirty, this is done always, since x uses dirty_lines even in
	   the nodirty methods */
	/* alloc lines + 1, dirty_lines must always end with empty line !! */
	dirty_lines=(struct dirty_line *)malloc(sizeof(struct dirty_line)*(visual_height+1));
	if (!dirty_lines)
	{
	   fprintf(stderr,"Cannot allocate space for dirtybuffer\n");
	   return OSD_NOT_OK;
	}
	memset((void *)dirty_lines, 0, sizeof(struct dirty_line)*(visual_height+1));
	
	/* secondary dirty buffer is only used when doing real dirty */
	if (use_dirty && !use_layer)
	{
	   /* alloc lines + 1, dirty_lines must always end with empty line !! */
	   old_dirty_lines=(struct dirty_line *)malloc(sizeof(struct dirty_line)*(visual_height+1));
	   if (!old_dirty_lines)
	   {
	      fprintf(stderr,"Cannot allocate space for dirtybuffer\n");
	      free (dirty_lines);
	      return OSD_NOT_OK;
	   }
	   memset((void *)old_dirty_lines, 0, sizeof(struct dirty_line)*(visual_height+1));
	}

	xsize = widthscale  * visual_width;
	ysize = heightscale * visual_height;
	
	/* Open the display. */

	display = XOpenDisplay (displayname);
	if(!display) {
		fprintf (stderr,"OSD ERROR: failed to open the display.\n");
		free(dirty_lines);
		if (use_dirty && !use_layer) free(old_dirty_lines);
		return OSD_NOT_OK;
	}

	screen 	 = DefaultScreenOfDisplay (display);
	myscreen = DefaultScreen(display);
	cursor   = XCreateFontCursor(display,XC_trek);

	/* test for available 8bit bitmaps resources */
	/* if not 8bit pseudocolor found, try to get any supported True color */
	/* set update display func properly */
	result 	 = XMatchVisualInfo(display,myscreen,8,PseudoColor,&myvisual);
	if ( ! result || force_truecolor ) {
	    force_truecolor=1;
	    fprintf(stderr,"8bit depth PseudoColor X-Visual not available :-( \n");
	    /* test for a 8bpp environment */
	    result = XMatchVisualInfo(display,myscreen,8,TrueColor,&myvisual);
	    if (result) { fprintf(stderr,"Using 8bpp TrueColor X-Visual Resource\n"); goto sigue;}
	    /* test for a 16bpp environment */
	    result = XMatchVisualInfo(display,myscreen,16,TrueColor,&myvisual);
	    if (result) { fprintf(stderr,"Using 16bpp TrueColor X-Visual Resource\n"); goto sigue;}
	    /* test for a 24bpp environment */
	    result = XMatchVisualInfo(display,myscreen,24,TrueColor,&myvisual);
	    if (result) { fprintf(stderr,"Using 24bpp TrueColor X-Visual Resource\n"); goto sigue;}
	    /* test for a 32bpp environment */
	    result = XMatchVisualInfo(display,myscreen,32,TrueColor,&myvisual);
	    if (result) { fprintf(stderr,"Using 32bpp TrueColor X-Visual Resource\n"); goto sigue;}
	    /*if arrives here means an error :-) */
	    fprintf(stderr,"Cannot find any supported X-Visual resource\n");
	    free(dirty_lines);
	    if (use_dirty && !use_layer) free(old_dirty_lines);
	    return OSD_NOT_OK;
	} else fprintf(stderr,"Using 8bpp PseudoColor Visual. Good!\n");
sigue:  /* bah */

	/* look for available Mitshm extensions */
	if (mit_shm_avail != -1 )
	{
	   mit_shm_avail = 0;
#ifdef USE_MITSHM
	   /* get XExtensions to be if mit shared memory is available */
	   result 	= XQueryExtension(display,"MIT-SHM",&i,&j,&k);
	   if ( ! result)
	      fprintf(stderr,"X-Server Doesn't support MIT-SHM extension\n");
	    else
	      mit_shm_avail = 1;
#endif
	} else mit_shm_avail = 0; /* has been user-disabled */

	/* now get a color map structure */
	if(force_truecolor && use_private_cmap)
	{
	  fprintf(stderr,"Private ColorMap use disabled in TrueColor mode\n");
	  use_private_cmap=0;
	}
	
	if(!use_private_cmap)
	{
	  colormap = DefaultColormapOfScreen(screen);
	  black_xpixel =  BlackPixelOfScreen(screen);
	}
	else
	{
	  XColor color;
	  colormap = XCreateColormap(display,RootWindowOfScreen(screen),myvisual.visual,AllocNone);
          color.flags = (DoRed | DoGreen | DoBlue);
	  color.pixel = 0;
	  color.red   = 0; 
	  color.green = 0; 
	  color.blue  = 0; 
	  if (XAllocColor (display,colormap,&color))
	  {
	    black_xpixel = color.pixel;
	  }
	  else
	  {
	    fprintf(stderr,"Couldn't allocate black_xpixel, this shouldn't happen!\nUsing 0\n");
	    black_xpixel = 0;
	  }
	  fprintf(stderr,"Using private color map\n");
	}
	
	/* Create the window. No buttons, no fancy stuff. */
	winattr.background_pixel      = black_xpixel;
	winattr.border_pixel          = WhitePixelOfScreen(screen);
	winattr.bit_gravity           = ForgetGravity;
	winattr.win_gravity           = NorthWestGravity;
	winattr.backing_store         = NotUseful;
	winattr.override_redirect     = False;
	winattr.save_under            = False;
	winattr.event_mask            = 0;
	winattr.do_not_propagate_mask = 0;
	winattr.colormap              = colormap;
	winattr.cursor                = None;
	window = XCreateWindow(display,RootWindowOfScreen(screen),0,0,xsize,ysize,0,myvisual.depth,
                           InputOutput,myvisual.visual,CWBorderPixel | CWBackPixel | CWBitGravity | \
                           CWWinGravity | CWBackingStore | CWOverrideRedirect | \
                           CWSaveUnder | CWEventMask | CWDontPropagate | \
                           CWColormap | CWCursor,&winattr);
  	if (!window) {
  		fprintf(stderr,"OSD ERROR: failed in XCreateWindow().\n");
		free(dirty_lines);
		if (use_dirty && !use_layer) free(old_dirty_lines);
  		return OSD_NOT_OK;
	}
	
	gc = XCreateGC(display,window,0,&xgcv);

	/*  Placement hints etc. */

	hints.flags	= PSize | PMinSize | PMaxSize;
 	hints.min_width	= hints.max_width = hints.base_width = xsize;
 	hints.min_height= hints.max_height = hints.base_height = ysize;

	wm_hints.input	= TRUE;
	wm_hints.flags	= InputHint;

	XSetWMHints 		(display, window, &wm_hints);
	XSetWMNormalHints 	(display, window, &hints);
	XStoreName 		(display, window, title);

	XDefineCursor(display,window,cursor);

#ifdef X11_JOYSTICK
	if(use_joystick) x11_joystick_init(); /* in devices.c */
#endif
	/* Map and expose the window. */
	if (use_mouse)
	{
	    /* grabb the pointer and query MotionNotify events */
	    XSelectInput(display, 
			 window, 
			 FocusChangeMask   | ExposureMask | 
			 KeyPressMask      | KeyReleaseMask | 
			 EnterWindowMask   | LeaveWindowMask |
			 PointerMotionMask | ButtonMotionMask |
			 ButtonPressMask   | ButtonReleaseMask
			);
	    XGrabPointer(display,
			 window, /* RootWindow(display,DefaultScreen(display)), */
			 False,
			 PointerMotionMask | ButtonMotionMask |
			 ButtonPressMask   | ButtonReleaseMask | 
			 EnterWindowMask   | LeaveWindowMask ,
			 GrabModeAsync, GrabModeAsync,
			 None, cursor, CurrentTime );
	} else {
	    XSelectInput(display, 
			 window, 
			 FocusChangeMask | ExposureMask | 
			 KeyPressMask | KeyReleaseMask
			 );
	}
	XMapRaised     (display, window);
	XClearWindow   (display, window);
	XWindowEvent   (display, window, ExposureMask, &event);

#ifdef USE_MITSHM
	if ( mit_shm_avail ) {
	    /* Create a MITSHM image. */

	    XSetErrorHandler( test_mit_shm );

	    /* CS: scaling changes requested image size */
	    image = XShmCreateImage (display, 
				myvisual.visual ,
				myvisual.depth,
       			    	ZPixmap, 
				NULL, 
				&shm_info, 
				xsize, 
				ysize);
  	    if (!image) {
  		fprintf (stderr,"OSD ERROR: failed in XShmCreateImage().\n");
		mit_shm_avail = 0;
		goto mit_shm_failed;
	    }
	    shm_info.shmid = shmget ( IPC_PRIVATE, 
				      image->bytes_per_line * image->height, 
				      (IPC_CREAT | 0777) );
	    if ( shm_info.shmid < 0) {
		    fprintf (stderr,"OSD ERROR: failed to create MITSHM block.\n");
		    free(dirty_lines);
		    if (use_dirty && !use_layer) free(old_dirty_lines);
		    return OSD_NOT_OK;
	    }

	    /* And allocate the bitmap buffer. */
            /* new pen color code force double buffering in every cases */
  	    scaled_buffer_ptr = image->data = shm_info.shmaddr = 
			(char *) shmat ( shm_info.shmid,0 ,0);
  	    if (!scaled_buffer_ptr) {
    	    	    fprintf (stderr,"OSD ERROR: failed to allocate MITSHM bitmap buffer.\n");
		    free(dirty_lines);
		    if (use_dirty && !use_layer) free(old_dirty_lines);
    		    return OSD_NOT_OK;
  	    }

	    shm_info.readOnly = FALSE;

	    /* Attach the MITSHM block. this will cause an exception if */
	    /* MIT-SHM is not available. so trap it and process         */

	    fprintf(stderr,"MIT-SHM Extension Available. trying to use...");

	    if(!XShmAttach (display, &shm_info)) {
  		    fprintf (stderr,"OSD ERROR: failed to attach MITSHM block.\n");
		    free(dirty_lines);
		    if (use_dirty && !use_layer) free(old_dirty_lines);
  		    return OSD_NOT_OK;
	    }
	    XSync(display,False); /* be sure to get request processed */
	    sleep(2); /* enought time to notify error if any */
#ifdef USE_TIMER
	    /* agh!! sleep() disables timer alarm. have to re-arm */
	    if (play_sound) start_timer();
#endif
	    XSetErrorHandler( None ); /* Restore error handler to default */
	    fprintf(stderr,"Success.\nUsing Shared Memory Features to speed up\n");

	} else { /* Not MITSHM Allocate a bitmap buffer. */
mit_shm_failed:  	/* bit dirty but usefull */
#endif
  	    scaled_buffer_ptr = (char *) malloc ( 4 * xsize * ysize );
  	    if (!scaled_buffer_ptr) {
    		fprintf (stderr,"OSD ERROR: failed to allocate bitmap buffer.\n");
		free(dirty_lines);
		if (use_dirty && !use_layer) free(old_dirty_lines);
    		return OSD_NOT_OK;
	    }
 	    /* CS: Ximage is allocated against our new image if scaling active */
  	    image = XCreateImage ( display, 
			myvisual.visual , 
			myvisual.depth, 
			ZPixmap, 
			0, 
			scaled_buffer_ptr,
			xsize, ysize,
			(myvisual.depth==8)?8:32, 
			0);
  	    if (!image) {
    		fprintf (stderr,"OSD ERROR: could not create image.\n");
    		free (scaled_buffer_ptr);
		free (dirty_lines);
		if (use_dirty && !use_layer) free(old_dirty_lines);
    		return OSD_NOT_OK;
  	    }
#ifdef USE_MITSHM
	}/* end of not MITSHM */
#endif
	/* call to routine to set update_display_func properly */
	if ( eval_update_display_func() == OSD_NOT_OK ) 
	{
	   if (mit_shm_avail == 0) free(scaled_buffer_ptr);
	   free (dirty_lines);
	   if (use_dirty && !use_layer) free(old_dirty_lines);
	   return OSD_NOT_OK;
	}

	/* set xpixel to black */
	for(i=0;i<256;i++) xpixel[i] = black_xpixel;
	/* set xpixel_allocated to FALSE */
	memset(xpixel_allocated, FALSE , sizeof(char) * 256);
	
	if (use_layer) fprintf(stderr,"Using layer system\n");
	if (use_dirty) fprintf(stderr,"Using dirty_buffer strategy\n");

	/* 
	   and now clear screen to black. ( should be done by default, but
	   some stupid servers has BlackWindowOfScreen set to White 
	*/
	XClearWindow(display,window);
	return OSD_OK;
}

/*
 * Shut down the display.
 */

void osd_close_display (void)
{
	int		i;

	if (display && window) {
		/* ungrab the pointer */
		if (use_mouse) XUngrabPointer(display,CurrentTime);

#ifdef USE_MITSHM
		if (mit_shm_avail) {
        	    XShmDetach (display, &shm_info);	/* Detach the SHM. */
      		    if (shm_info.shmaddr) shmdt (shm_info.shmaddr);
      		    if (shm_info.shmid >= 0) shmctl (shm_info.shmid, IPC_RMID, 0);
		}
#endif
		/* free the bitmap & dirty_lines */
		osd_free_bitmap(bitmap);
		free(dirty_lines);
		if (use_dirty && !use_layer) free(old_dirty_lines);

		/* Destroy the image and free the colors. */
		if (image) {
    			XDestroyImage (image);
   		}
    		for (i = 0; i < 256; i++) {
			if (xpixel_allocated[i])
      			   XFreeColors (display, colormap, &xpixel[i], 1, 0);
  		}
		if (use_private_cmap)
		{
      			XFreeColors  (display, colormap, &black_xpixel, 1, 0);
			XFreeColormap(display, colormap);
		}
  		if (display) {
  			XAutoRepeatOn (display);
  			XFlush (display); /* send all events to sync; */
  			XCloseDisplay (display);
  		}
	}
}


/*
 * Set the screen colors using the given palette.
 *
 */
void sysdep_init_palette(int totalcolors,const unsigned char *palette, unsigned short *pens)
{
       XColor color;
       int i,j;
       
       if (totalcolors==256) j=0; else j=1;
       
       /* and allocate color cells */ 
       for (i=0; i < totalcolors; i++) {
          color.flags = (DoRed | DoGreen | DoBlue);
	  color.pixel = i+j;
	  color.red   = (int)palette[3*i]   << 8; 
	  color.green = (int)palette[3*i+1] << 8; 
	  color.blue  = (int)palette[3*i+2] << 8; 
	  
          current_palette[(i+j)*3   ] = palette[3*i  ];
          current_palette[(i+j)*3 +1] = palette[3*i+1];
          current_palette[(i+j)*3 +2] = palette[3*i+2];
          
	  if (XAllocColor (display,colormap,&color))
	  {
	    xpixel[i+j] = color.pixel;
	    xpixel_allocated[i+j] = TRUE;
	  }
	  
          pens[i] = i+j;
       }
}

/* not needed under X */
void sysdep_clear_screen(void)
{
}

void osd_modify_pen(int pen, unsigned char red,unsigned char green,unsigned char blue) 
{
	XColor		color;
	
	/* Optimise out operation if no palette change is required */
	if (current_palette[pen*3   ] == red &&
	    current_palette[pen*3 +1] == green &&
	    current_palette[pen*3 +2] == blue)
	    return;
	
	/* free previously allocated color */
	if (xpixel_allocated[pen])
	{
	   XFreeColors (display, colormap, &xpixel[pen], 1, 0);
	   xpixel_allocated[pen] = FALSE;
	}
	   
	/* Translate VGA 0-63 values of new color to X 0-65536 values. */
	color.flags	= (DoRed | DoGreen | DoBlue);
	color.red	= (int) red << 8;
	color.green	= (int) green << 8;
	color.blue	= (int) blue << 8;
	
	current_palette[pen*3   ] = red;
	current_palette[pen*3 +1] = green;
	current_palette[pen*3 +2] = blue;
	
	/* allocate new color and assign it to pen index */
	if (XAllocColor (display,colormap,&color)) 
	{
		if (xpixel[pen] != color.pixel)
		   palette_dirty   = TRUE;
		if (pen         == black_pen  )
		   black_pen_valid = FALSE;
  		xpixel[pen]               = color.pixel;
		xpixel_allocated[pen]     = TRUE ;
	} else {
		/* If color allocation failed, use black to ensure the
		   pen is not left set to an invalid color */
		xpixel[pen] = black_xpixel;
	}
}

#endif /* ndef svgalib */
#endif /* ndef ggi */
